@@ -1,4 +1,4 @@
-# copyright 2004-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
+# copyright 2004-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
 # contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
 #
 # This file is part of yams.
@@ -15,9 +15,8 @@
 #
 # You should have received a copy of the GNU Lesser General Public License along
 # with yams. If not, see <http://www.gnu.org/licenses/>.
-"""Exceptions shared by different ER-Schema modules.
+"""YAMS exception classes"""
 
-"""
 __docformat__ = "restructuredtext en"
 
 class SchemaError(Exception):
@@ -66,26 +65,65 @@ class BadSchemaDefinition(SchemaError):
 
 
 class ValidationError(SchemaError):
-    """validation error details the reason(s) why the validation failed
+    """Validation error details the reason(s) why the validation failed.
 
-    :type entity: EntityType
-    :param entity: the entity that could not be validated
+    Arguments are:
 
-    :type explanation: dict
-    :param explanation: pairs of (attribute, error)
+    * `entity`: the entity that could not be validated; actual type depends on
+      the client library
+
+    * `errors`: errors dictionary, None key used for global error, other keys
+      should be attribute/relation of the entity, qualified as subject/object
+      using :func:`yams.role_name`.  Values are the message associated to the
+      keys, and may include interpolation string starting with '%(KEY-' where
+      'KEY' will be replaced by the associated key once the message has been
+      translated. This allows predictable/translatable message and avoid args
+      conflict if used for several keys.
+
+    * `msgargs`: dictionary of substitutions to be inserted in error
+      messages once translated (only if msgargs is given)
+
+    * `i18nvalues`: list of keys in msgargs whose value should be translated
+
+    Translation will be done **in-place** by calling :meth:`translate`.
     """
 
-    def __init__(self, entity, explanation):
+    def __init__(self, entity, errors, msgargs=None, i18nvalues=None):
         # set args so ValidationError are serializable through pyro
-        SchemaError.__init__(self, entity, explanation)
+        SchemaError.__init__(self, entity, errors)
         self.entity = entity
-        assert isinstance(explanation, dict), \
-            'validation error explanation must be a dict'
-        self.errors = explanation
+        assert isinstance(errors, dict), 'validation errors must be a dict'
+        self.errors = errors
+        self.msgargs = msgargs
+        self.i18nvalues = i18nvalues
+        self._translated = False
 
     def __unicode__(self):
+        if not self._translated:
+            self.translate(unicode)
         if len(self.errors) == 1:
             attr, error = self.errors.items()[0]
             return u'%s (%s): %s' % (self.entity, attr, error)
         errors = '\n'.join('* %s: %s' % (k, v) for k, v in self.errors.items())
         return u'%s:\n%s' % (self.entity, errors)
+
+    def translate(self, _):
+        """Translate and interpolate messsages in the errors dictionary, using
+        the given translation function.
+
+        If no substitution has been given, suppose msg is already translated for
+        bw compat, so no translation occurs.
+
+        This method may only be called once.
+        """
+        assert not self._translated
+        self._translated = True
+        if self.msgargs is not None:
+            if self.i18nvalues:
+                for key in self.i18nvalues:
+                    self.msgargs[key] = _(self.msgargs[key])
+            for key, msg in self.errors.iteritems():
+                msg = _(msg).replace('%(KEY-', '%('+key+'-')
+                self.errors[key] = msg % self.msgargs
+
+
